/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.BiFunction;
import com.github.sviperll.BinaryOperator;
import com.github.sviperll.Function;
import com.github.sviperll.OptionalVisitor;
import com.github.sviperll.Supplier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
public class Collectors {
    public static <T, U, R, E extends Exception> Collector<T, R, E> optional(final Function<? super T, CollectorState<? super T, U, E>> notOptionalCollector, final OptionalVisitor<? super U, R, E> visitor) {
        return new Collector<T, R, E>(new Supplier<CollectorState<T, R, E>>() {
            @Override
            public CollectorState<T, R, E> get() {
                StatefulCollectorState<T, R, E> state = new StatefulCollectorState<T, R, E>(null);
                NoResultCollectorState<T, U, R, E> noResultCollectorState = new NoResultCollectorState<T, U, R, E>(visitor, state, notOptionalCollector);
                state.setBehaviour(noResultCollectorState);
                return state;
            }
        });
    }

    public static <T, U> Collector<T, U, RuntimeException> reducing(final U seed, final BiFunction<U, ? super T, U> function) {
        return new Collector<T, U, RuntimeException>(new Supplier<CollectorState<T, U, RuntimeException>>() {
            @Override
            public CollectorState<T, U, RuntimeException> get() {
                return new ReducingCollectorState<T, U, RuntimeException>(seed, function);
            }
        });
    }

    public static <T, R, E extends Exception> Collector<T, R, E> reducing(final BinaryOperator<T> operator, final OptionalVisitor<? super T, R, E> visitor) {
        Function<T, CollectorState<? super T, T, E>> stateFactory = new Function<T, CollectorState<? super T, T, E>>() {
            @Override
            public CollectorState<T, T, E> apply(T seed) {
                return new ReducingCollectorState<T, T, E>(seed, operator);
            }
        };
        return optional(stateFactory, visitor);
    }

    public static <T> Collector<T, List<T>, RuntimeException> toList() {
        return toCollection(new Supplier<List<T>>() {
            @Override
            public List<T> get() {
                return new ArrayList<T>();
            }
        });
    }

    public static <T> Collector<T, HashSet<T>, RuntimeException> toHashSet() {
        return toCollection(new Supplier<HashSet<T>>() {
            @Override
            public HashSet<T> get() {
                return new HashSet<T>();
            }
        });
    }

    public static <T> Collector<T, TreeSet<T>, RuntimeException> toTreeSet() {
        return toCollection(new Supplier<TreeSet<T>>() {
            @Override
            public TreeSet<T> get() {
                return new TreeSet<T>();
            }
        });
    }

    static <T, R, E extends Exception> Collector<T, R, E> findFirst(OptionalVisitor<? super T, R, E> visitor) {
        Function<T, CollectorState<? super T, T, E>> stateFactory = new Function<T, CollectorState<? super T, T, E>>() {
            @Override
            public CollectorState<T, T, E> apply(final T firstElement) {
                return new CollectorState<T, T, E>() {
                    @Override
                    public T get() throws E {
                        return firstElement;
                    }

                    @Override
                    public void accept(T value) {
                    }

                    @Override
                    public boolean needsMore() {
                        return false;
                    }
                };
            }
        };
        return optional(stateFactory, visitor);
    }

    public static <T, R extends Collection<T>> Collector<T, R, RuntimeException> toCollection(final Supplier<R> factory) {
        return new Collector<T, R, RuntimeException>(new Supplier<CollectorState<T, R, RuntimeException>>() {
            @Override
            public CollectorState<T, R, RuntimeException> get() {
                return new CollectorState<T, R, RuntimeException>() {
                    R result = factory.get();
                    @Override
                    public R get() throws RuntimeException {
                        return result;
                    }

                    @Override
                    public void accept(T value) {
                        result.add(value);
                    }

                    @Override
                    public boolean needsMore() {
                        return true;
                    }
                };
            }
        });
    }

    public static <I, K, V, R extends Map<K, V>> Collector<I, R, RuntimeException> toMap(final Function<? super I, ? extends K> function, final Supplier<? extends CollectorState<I, V, ? extends RuntimeException>> collector, final Supplier<R> factory) {
        return new Collector<I, R, RuntimeException>(new Supplier<CollectorState<I, R, RuntimeException>>() {
            @Override
            public CollectorState<I, R, RuntimeException> get() {
                return new CollectorState<I, R, RuntimeException>() {
                    HashMap<K, CollectorState<I, V, ? extends RuntimeException>> state = new HashMap<K, CollectorState<I, V, ? extends RuntimeException>>();
                    @Override
                    public R get() throws RuntimeException {
                        R result = factory.get();
                        for (Map.Entry<K, CollectorState<I, V, ? extends RuntimeException>> entry: state.entrySet()) {
                            result.put(entry.getKey(), entry.getValue().get());
                        }
                        return result;
                    }

                    @Override
                    public void accept(I value) {
                        K key = function.apply(value);
                        CollectorState<I, V, ? extends RuntimeException> s = state.get(key);
                        if (s == null) {
                            s = collector.get();
                            state.put(key, s);
                        }
                        s.accept(value);
                    }

                    @Override
                    public boolean needsMore() {
                        return true;
                    }
                };
            }
        });
    }

    public static <I, K, V> Collector<I, HashMap<K, V>, RuntimeException> toHashMap(final Function<? super I, ? extends K> function, final Supplier<? extends CollectorState<I, V, ? extends RuntimeException>> collector) {
        return toMap(function, collector, new Supplier<HashMap<K, V>>() {
            @Override
            public HashMap<K, V> get() {
                return new HashMap<K, V>();
            }
        });
    }

    public static <I, K, V> Collector<I, TreeMap<K, V>, RuntimeException> toTreeMap(final Function<? super I, ? extends K> function, final Supplier<? extends CollectorState<I, V, ? extends RuntimeException>> collector) {
        return toMap(function, collector, new Supplier<TreeMap<K, V>>() {
            @Override
            public TreeMap<K, V> get() {
                return new TreeMap<K, V>();
            }
        });
    }

    public static <T> Collector<T, Integer, RuntimeException> counting() {
        return new Collector<T, Integer, RuntimeException>(new Supplier<CollectorState<T, Integer, RuntimeException>>() {
            @Override
            public CollectorState<T, Integer, RuntimeException> get() {
                return new CollectorState<T, Integer, RuntimeException>() {
                    private int count = 0;

                    @Override
                    public void accept(T value) {
                        count++;
                    }

                    @Override
                    public boolean needsMore() {
                        return true;
                    }

                    @Override
                    public Integer get() throws RuntimeException {
                        return count;
                    }
                };
            }
        });
    }

    private static class StatefulCollectorState<T, R, E extends Exception> implements CollectorState<T, R, E> {
        private CollectorState<T, R, E> behaviour;

        public StatefulCollectorState(CollectorState<T, R, E> behaviour) {
            this.behaviour = behaviour;
        }

        public void setBehaviour(CollectorState<T, R, E> behaviour) {
            this.behaviour = behaviour;
        }

        @Override
        public void accept(T value) {
            behaviour.accept(value);
        }

        @Override
        public R get() throws E {
            return behaviour.get();
        }

        @Override
        public boolean needsMore() {
            return behaviour.needsMore();
        }
    }

    private static class NoResultCollectorState<T, U, R, E extends Exception> implements CollectorState<T, R, E> {
        private final OptionalVisitor<? super U, R, E> visitor;
        private final StatefulCollectorState<T, R, E> state;
        private final Function<? super T, CollectorState<? super T, U, E>> collector;
        public NoResultCollectorState(OptionalVisitor<? super U, R, E> visitor, StatefulCollectorState<T, R, E> state, Function<? super T, CollectorState<? super T, U, E>> collector) {
            this.visitor = visitor;
            this.state = state;
            this.collector = collector;
        }

        @Override
        public R get() throws E {
            return visitor.missing();
        }

        @Override
        public void accept(T value) {
            CollectorState<? super T, U, E> reducingConsumer = collector.apply(value);
            state.setBehaviour(new NotOptionalCollectorState<T, U, R, E>(reducingConsumer, visitor));
        }

        @Override
        public boolean needsMore() {
            return true;
        }
    }

    private static class ReducingCollectorState<T, U, E extends Exception> implements CollectorState<T, U, E> {
        private final BiFunction<U, ? super T, U> function;
        private U result;

        public ReducingCollectorState(U seed,
                                BiFunction<U, ? super T, U> function) {
            this.result = seed;
            this.function = function;
        }

        @Override
        public void accept(T value) {
            result = function.apply(result, value);
        }

        @Override
        public U get() {
            return result;
        }

        @Override
        public boolean needsMore() {
            return true;
        }
    }

    private static class NotOptionalCollectorState<T, U, R, E extends Exception> implements CollectorState<T, R, E> {
        private final CollectorState<? super T, U, E> consumer;
        private final OptionalVisitor<? super U, R, E> visitor;

        public NotOptionalCollectorState(CollectorState<? super T, U, E> consumer, OptionalVisitor<? super U, R, E> visitor) {
            this.consumer = consumer;
            this.visitor = visitor;
        }

        @Override
        public R get() throws E {
            return visitor.present(consumer.get());
        }

        @Override
        public void accept(T value) {
            consumer.accept(value);
        }

        @Override
        public boolean needsMore() {
            return consumer.needsMore();
        }
    }
}
