/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.InterruptableConsumer;
import com.github.sviperll.Consumer;
import com.github.sviperll.io.Closeables;
import java.io.Closeable;
import java.util.Collections;
import java.util.Iterator;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
class FlattenedStreamBehaviour<T> implements CloseableStreamBehaviour<T> {
    private final Stream<? extends Stream<T>> stream;
    public FlattenedStreamBehaviour(Stream<? extends Stream<T>> stream) {
        this.stream = stream;
    }

    @Override
    public CloseableIterator<T> openIterator() {
        Owned<CloseableIterator<? extends Stream<T>>> outerCloseable = Owned.<CloseableIterator<? extends Stream<T>>>newInstance(stream.openIterator());
        try {
            Iterator<? extends Stream<T>> outerIterator = outerCloseable.value();
            if (!outerIterator.hasNext())
                return new CloseableIterator<T>(Collections.<T>emptyIterator(), Closeables.empty());
            else {
                Owned<CloseableIterator<T>> innerCloseable = Owned.newInstance(outerIterator.next().openIterator());
                try {
                    Iterator<T> innerIterator = innerCloseable.value();
                    while (!innerIterator.hasNext()) {
                        if (!outerIterator.hasNext())
                            return new CloseableIterator<T>(Collections.<T>emptyIterator(), Closeables.empty());
                        else {
                            innerCloseable.set(outerIterator.next().openIterator());
                            innerIterator = innerCloseable.value();
                        }
                    }
                    T firstElement = innerIterator.next();
                    while (!innerIterator.hasNext() && outerIterator.hasNext()) {
                        innerCloseable.set(outerIterator.next().openIterator());
                        innerIterator = innerCloseable.value();
                    }
                    return FlattenedIterator.newInstance(Owned.newInstance(innerCloseable.take()), firstElement, outerCloseable.take());
                } finally {
                    innerCloseable.close();
                }
            }
        } finally {
            outerCloseable.close();
        }
    }

    @Override
    public void forEach(final InterruptableConsumer<? super T> consumer) {
        stream.forEach(new InterruptableConsumer<Stream<T>>() {
            @Override
            public void accept(Stream<T> value) {
                value.forEach(consumer);
            }

            @Override
            public boolean needsMore() {
                return consumer.needsMore();
            }
        });
    }

    private static class FlattenedIterator<T> implements Iterator<T>, Closeable {
        public static <T> CloseableIterator<T> newInstance(Owned<CloseableIterator<T>> innerCloseable,
                                                           T firstElement,
                                                           CloseableIterator<? extends Stream<T>> outerIterator) {
            FlattenedIterator<T> flattenedIterator = new FlattenedIterator<T>(innerCloseable, innerCloseable.value(), firstElement, outerIterator);
            return new CloseableIterator<T>(flattenedIterator, flattenedIterator);
        }

        private final CloseableIterator<? extends Stream<T>> outerIterator;
        private final Owned<CloseableIterator<T>> innerCloseable;
        private CloseableIterator<T> innerIterator;
        private T nextElement;
        private boolean hasNext = true;

        private FlattenedIterator(Owned<CloseableIterator<T>> innerCloseable,
                                  CloseableIterator<T> innerIterator,
                                  T firstElement,
                                  CloseableIterator<? extends Stream<T>> outerIterator) {
            this.innerCloseable = innerCloseable;
            this.nextElement = firstElement;
            this.outerIterator = outerIterator;
            this.innerIterator = innerIterator;
        }

        @Override
        public void close() {
            try {
                innerCloseable.close();
            } finally {
                outerIterator.close();
            }
        }

        @Override
        public boolean hasNext() {
            return hasNext;
        }

        @Override
        public T next() {
            T result = nextElement;
            while (!innerIterator.hasNext()) {
                if (!outerIterator.hasNext()) {
                    hasNext = false;
                    return result;
                } else {
                    innerCloseable.set(outerIterator.next().openIterator());
                    innerIterator = innerCloseable.value();
                }
            }
            nextElement = innerIterator.next();
            return result;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("Not supported.");
        }
    }

}
