/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.InterruptableConsumer;
import com.github.sviperll.Consumer;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
class LimitedStreamBehaviour<T> implements CloseableStreamBehaviour<T> {
    private final CloseableStreamBehaviour<T> stream;
    private final int size;

    public LimitedStreamBehaviour(CloseableStreamBehaviour<T> stream, int size) {
        this.stream = stream;
        this.size = size;
    }

    @Override
    public CloseableIterator<T> openIterator() {
        Owned<CloseableIterator<T>> closeableIterator = Owned.newInstance(stream.openIterator());
        try {
            final Iterator<T> iterator = closeableIterator.value();
            Iterator<T> limitedIterator = new Iterator<T>() {
                private int count = 0;

                @Override
                public boolean hasNext() {
                    return iterator.hasNext() && count < size;
                }

                @Override
                public T next() {
                    if (!hasNext())
                        throw new NoSuchElementException();
                    T result = iterator.next();
                    count++;
                    return result;
                }

                @Override
                public void remove() {
                    throw new UnsupportedOperationException("Not supported.");
                }
            };
            return new CloseableIterator<T>(limitedIterator, closeableIterator.take());
        } finally {
            closeableIterator.close();
        }
    }

    @Override
    public void forEach(final InterruptableConsumer<? super T> consumer) {
        stream.forEach(new InterruptableConsumer<T>() {
            int count = 0;
            @Override
            public void accept(T value) {
                if (count < size)
                    consumer.accept(value);
                count++;
            }

            @Override
            public boolean needsMore() {
                return count < size && consumer.needsMore();
            }
        });
    }
}
