/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.InterruptableConsumer;
import com.github.sviperll.Consumer;
import com.github.sviperll.Supplier;
import com.github.sviperll.io.Closeables;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.nio.charset.Charset;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
public class StreamableTextFile {
    private final File file;
    private final Charset charset;

    public StreamableTextFile(File file, Charset charset) {
        this.file = file;
        this.charset = charset;
    }

    public File file() {
        return file;
    }

    public StreamableFile binary() {
        return new StreamableFile(file);
    }

    public Stream<String> lines() {
        return Stream.newInstance(new Supplier<CloseableStream<String>>() {
            @Override
            public CloseableStream<String> get() {
                try {
                    Owned<BufferedInputStream> closeableInputStream = Owned.newInstance(new BufferedInputStream(new FileInputStream(file)));
                    try {
                        Owned<BufferedReader> closeableReader = Owned.newInstance(new BufferedReader(new InputStreamReader(closeableInputStream.take(), charset)));
                        try {
                            BufferedReader reader = closeableReader.value();
                            String firstLine = reader.readLine();
                            if (firstLine == null)
                                return new CloseableStream(Stream.<Byte>empty(), Closeables.empty());
                            else {
                                Stream<String> stream = Stream.newInstance(new LineReaderStream(firstLine, reader));
                                return new CloseableStream(stream, closeableReader.take());
                            }
                        } finally {
                            closeableReader.close();
                        }
                    } finally {
                        closeableInputStream.close();
                    }
                } catch (IOException ex) {
                    throw new RuntimeException(ex);
                }
            }
        });
    }

    public Stream<Character> characters() {
        return Stream.newInstance(new Supplier<CloseableStream<Character>>() {
            @Override
            public CloseableStream<Character> get() {
                try {
                    Owned<BufferedInputStream> closeableInputStream = Owned.newInstance(new BufferedInputStream(new FileInputStream(file)));
                    try {
                        Owned<BufferedReader> closeableReader = Owned.newInstance(new BufferedReader(new InputStreamReader(closeableInputStream.take(), charset)));
                        try {
                            BufferedReader reader = closeableReader.value();
                            int firstCharacter = reader.read();
                            if (firstCharacter < 0)
                                return new CloseableStream(Stream.<Byte>empty(), Closeables.empty());
                            else {
                                Stream<Character> stream = Stream.newInstance(new CharacterReaderStream((char)firstCharacter, reader));
                                return new CloseableStream(stream, closeableReader.take());
                            }
                        } finally {
                            closeableReader.close();
                        }
                    } finally {
                        closeableInputStream.close();
                    }
                } catch (IOException ex) {
                    throw new RuntimeException(ex);
                }
            }
        });
    }

    private static class LineReaderStream implements StreamBehaviour<String> {
        private final String firstLine;
        private final BufferedReader reader;

        private LineReaderStream(String firstLine, BufferedReader reader) {
            this.firstLine = firstLine;
            this.reader = reader;
        }
        @Override
        public Iterator<String> openIterator() {
            return new LineBufferedReaderIterator(firstLine, reader);
        }

        @Override
        public void forEach(InterruptableConsumer<? super String> consumer) {
            if (consumer.needsMore()) {
                consumer.accept(firstLine);
                String line;
                try {
                    while ((line = reader.readLine()) != null) {
                        if (!consumer.needsMore())
                            break;
                        consumer.accept(line);
                    }
                } catch (IOException ex) {
                    throw new RuntimeException(ex);
                }
            }
        }
    }

    private static class LineBufferedReaderIterator implements Iterator<String> {
        private final BufferedReader bufferedReader;
        private boolean hasNext = true;
        private String nextLine;

        public LineBufferedReaderIterator(String firstLine, BufferedReader bufferedReader) {
            this.nextLine = firstLine;
            this.bufferedReader = bufferedReader;
        }

        @Override
        public boolean hasNext() {
            return hasNext;
        }

        @Override
        public String next() {
            if (!hasNext)
                throw new NoSuchElementException();
            String result = nextLine;
            try {
                nextLine = bufferedReader.readLine();
            } catch (IOException ex) {
                throw new RuntimeException(ex);
            }
            hasNext = nextLine != null;
            return result;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("Not supported.");
        }

    }

    private static class CharacterReaderStream implements StreamBehaviour<Character> {
        private final char firstCharacter;
        private final Reader reader;

        private CharacterReaderStream(char firstLine, Reader reader) {
            this.firstCharacter = firstLine;
            this.reader = reader;
        }
        @Override
        public Iterator<Character> openIterator() {
            return new CharacterBufferedReaderIterator(firstCharacter, reader);
        }

        @Override
        public void forEach(InterruptableConsumer<? super Character> consumer) {
            if (consumer.needsMore()) {
                consumer.accept(firstCharacter);
                int character;
                try {
                    while ((character = reader.read()) >= 0) {
                        if (!consumer.needsMore())
                            break;
                        consumer.accept((char)character);
                    }
                } catch (IOException ex) {
                    throw new RuntimeException(ex);
                }
            }
        }
    }

    private static class CharacterBufferedReaderIterator implements Iterator<Character> {
        private final Reader reader;
        private char nextCharacter;
        private boolean hasNext = true;

        public CharacterBufferedReaderIterator(char firstCharacter, Reader reader) {
            this.nextCharacter = firstCharacter;
            this.reader = reader;
        }

        @Override
        public boolean hasNext() {
            return hasNext;
        }

        @Override
        public Character next() {
            if (!hasNext)
                throw new NoSuchElementException();
            char result = nextCharacter;
            int character;
            try {
                character = reader.read();
            } catch (IOException ex) {
                throw new RuntimeException(ex);
            }
            if (character < 0)
                hasNext = false;
            else
                nextCharacter = (char)character;
            return result;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("Not supported.");
        }
    }
}
